/* 
 * vo_file.c - MPlayer video out into a file.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * THIS  SOFTWARE  IS PROVIDED	  ``AS	IS'' AND   ANY	EXPRESS OR IMPLIED
 * WARRANTIES,	  INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 * NO	EVENT  SHALL   THE AUTHOR  BE	 LIABLE FOR ANY	  DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED	  TO, PROCUREMENT OF  SUBSTITUTE GOODS	OR SERVICES; LOSS OF
 * USE, DATA,	OR PROFITS; OR	BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN	 CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * You should have received a copy of the  GNU General Public License along
 * with this program; if not, write  to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Copyright 2005 (C) ARC International.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <linux/fb.h>

#include "config.h"
#include "fastmemcpy.h"
#include "video_out.h"
#include "video_out_internal.h"
#include "aspect.h"
#include "sub.h"

static vo_info_t info =
{
	"Dummy",
	"dummy",
	"",
	""
};

#define SCREEN_WIDTH 704
#define SCREEN_HEIGHT 480
#define YLEN SCREEN_WIDTH*SCREEN_HEIGHT
#define UVLEN (SCREEN_WIDTH/2)*(SCREEN_HEIGHT/2)

LIBVO_EXTERN(dummy)

typedef struct {
  uint8_t Y[YLEN];
  uint8_t U[UVLEN];
  uint8_t V[UVLEN];
} YUV_plane;

static int framecount;
static void (*alpha_func)(), (*alpha_func_double)();
static YUV_plane outbuf;
static int verbose = 0;

static uint32_t preinit(const char *arg)
{
	if (verbose) printf("vo file: preinit.\n");
	framecount = 0;

	return 0;
}


static void uninit(void)
{
	if (verbose) printf("vo file: unint.\n");
}


static void clear_screen()
{
		if (verbose) printf("vo file: clear_screen.\n");
}


static void setup_screen(uint32_t full)
{
		if (verbose) printf("vo file: setup_screen.\n");
}


static uint32_t config(uint32_t width, uint32_t height, uint32_t d_width, uint32_t d_height,
		uint32_t flags, char *title, uint32_t format)
{
	if (verbose) printf("vo file: config.\n");
	// This lot isn't
	alpha_func = 0;
	alpha_func_double = 0;

	switch(format) {
	  case IMGFMT_YV12:
	  case IMGFMT_I420:
	  case IMGFMT_IYUV:
			break;

	  default:
			printf("vo file: Only YUV4:2:0 supported.\n");
			return -1;
	}

	return 0;
}


/* Double-buffering draw_alpha */
static void draw_alpha_double(int x, int y, int w, int h, unsigned char *src,
		unsigned char *srca, int stride)
{
		if (verbose) printf("vo file: draw_alpha_double.\n");
}


/* Single-buffering draw_alpha */
static void draw_alpha(int x, int y, int w, int h, unsigned char *src,
		unsigned char *srca, int stride)
{
		if (verbose) printf("vo file:draw_alpha.\n");
}


static void draw_osd(void)
{
		if (verbose) printf("vo file:draw_osd.\n");
}


static void flip_page(void)
{
  // This is called when a screen draw has been completed, so
  // this is where we can save a frame to disk, calculate its
  // CRC or whatever.
  framecount++;
}


static uint32_t draw_frame(uint8_t *src[])
{
  // No RGB output.
  if (verbose) printf("vo_file:draw_frame.\n");
  return 0;
}


static uint32_t draw_slice(uint8_t *src[], int stride[], int w, int h, int x, int y)
{
	// Renders a slice of data into the 'display' buffer. Can be called multiple times
  // completely render a frame
	int i, y_offset;
	void *dst;
	void *srcp;

	//		printf("vo file: draw_slice entered w %d h %d x %d y %d long offset %d short offset %d uBase0 0x%x\n",w,h,x,y,y*320,y*640,(uBase0 + (y*320)));
	
	y_offset = y * SCREEN_WIDTH + x;
	srcp = src[0];
	dst=outbuf.Y + y_offset;
	for (i=0;i<h;i++) {
		memcpy(dst,srcp,w);
		dst += SCREEN_WIDTH;
		srcp += stride[0];
	}

	srcp = src[1];
	y_offset = (y/2) * (SCREEN_WIDTH/2) + x/2;
	w = w/2;
  h = h/2;
	dst=outbuf.U + y_offset;
	for (i=0;i<h;i++) {
		memcpy(dst,srcp,w);
		dst += (SCREEN_WIDTH/2);
		srcp += stride[1];
	}

	srcp = src[2];
	dst=outbuf.V + y_offset;
	for (i=0;i<h;i++) {
		memcpy(dst,srcp,w);
		dst += (SCREEN_WIDTH/2);
		srcp += stride[2];
	}

	return 0;
}


static uint32_t get_image(mp_image_t *mpi)
{
	// No direct rendering here
  if (verbose) printf("vo file: get_image.\n");
	mpi->planes[0] = outbuf.Y;
	mpi->planes[1] = outbuf.U;
	mpi->planes[2] = outbuf.V;
	mpi->flags |= MP_IMGFLAG_DIRECT;
  return VO_TRUE;
}


static uint32_t control(uint32_t request, void *data, ...)
{
	if (verbose) printf("vo file:control.\n");
  switch(request) {
  case VOCTRL_GET_IMAGE:
    return get_image(data);
    
  case VOCTRL_DRAW_IMAGE:
    return VO_TRUE;
    
  case VOCTRL_QUERY_FORMAT:
    switch(*((uint32_t*)data)) {
    case IMGFMT_YV12:
    case IMGFMT_I420:
    case IMGFMT_IYUV:
			return VFCAP_CSP_SUPPORTED | VFCAP_CSP_SUPPORTED_BY_HW;
		}

		return 0;		/* Format not supported */

	case VOCTRL_FULLSCREEN:
		return 0;

	}

	return VO_NOTIMPL;
}


static void check_events(void) {}
